#pragma once

#include <memory>
#include <type_traits>
#include <cassert>

template< typename T, size_t _block_size = 256  >
struct link_n_list
{
    using value_type = T;
    using reference = T &;
    using const_reference = const T &;
    using difference_type = ptrdiff_t;
    using size_type = size_t;

    static constexpr size_t block_size = _block_size;

  private:
    struct Node
    {
        Node() : _end( begin() ) { }
        Node( const Node &o, Node *prev ) :
            _end( begin() + o.size() ), _prev( prev )
        {
            std::uninitialized_copy( o.begin(), o.end(), begin() );
        }
        ~Node() {
            std::destroy( begin(), end() );
        }

        size_type size() const { return end() - begin(); }

        bool full() const {
            return _end == begin() + block_size;
        }

        bool empty() const {
            return _end == begin();
        }

        // note: reinterpret_cast from char pointer is one of the few OK ones
        value_type *begin() { return reinterpret_cast< value_type * >( _data ); } // NOLINT
        const value_type *begin() const { return reinterpret_cast< const value_type * >( _data ); } // NOLINT

        value_type *end() { return _end; }
        const value_type *end() const { return _end; }

        template< typename TT >
        reference push_back( TT &&x ) {
            assert( !full() );
            // inplace new: construct element in given, already allocated, space
            new ( _end ) value_type( std::forward< TT >( x ) );
            ++_end;
            return _end[-1];
        }

        // forward pointers are unique_ptr, back has to be raw to prevent
        // circular structure of unique_ptrs
        std::unique_ptr< Node > _next;
        value_type *_end;
        // alignas( alignof( value_type ) ) ensures data are properly aligned
        alignas( alignof( value_type ) ) char _data[ block_size * sizeof( value_type ) ];
        Node *_prev = nullptr;
    };
  public:

    // fill in forward iterators
    struct iterator;
    struct const_iterator;

    link_n_list() : _first( std::make_unique< Node >() ), _last( _first.get() ) { }
    // no need for explicit destructor as unique_ptrs take care of deallocation

    // copying is tricky with unique_ptrs, we have to build new linked list
    // based on the original, copying nodes is no good (and should not be
    // allowed)
    link_n_list( const link_n_list &o ) {
        if ( o.empty() )
            return;

        const Node *ptr = o._first.get();
        _first = std::make_unique< Node >( *ptr, nullptr );
        Node *prev = _first.get();
        ptr = ptr->_next.get();

        while ( ptr ) {
            prev->_next = std::make_unique< Node >( *ptr, prev );
            prev = prev->_next.get();
            ptr = ptr->_next.get();
        }
        assert( prev );
        _last = prev;
    }
    link_n_list( link_n_list &&o ) noexcept = default;

    link_n_list &operator=( link_n_list o ) {
        swap(o);
    }

    void swap( link_n_list &o ) {
        using std::swap;
        swap( _first, o._first );
        swap( _last, o._last );
    }

    // unique_ptr has only explicit cast to bool, so either compare it with
    // nullptr, or use explicit cast such as bool( _first )
    bool empty() const { return _first->empty(); }

    iterator begin();
    const_iterator begin() const;
    const_iterator cbegin() const;

    iterator end();
    const_iterator end() const;
    const_iterator cend() const;

    reference front() { return _first->begin()[0]; }
    const_reference front() const { return _first->begin()[0]; }

    reference back() { return _last->end()[-1]; }
    const_reference back() const { return _last->end()[-1]; }

    // iterators are never invalidated by push_back
    reference push_back( const value_type &v ) { return _push_back( v ); }
    reference push_back( value_type &&v ) { return _push_back( std::move( v ) ); }

  private:
    template< typename TT >
    reference _push_back( TT &&val ) {
        if ( !_last ) {
            _first = std::make_unique< Node >();
            _last = _first.get();
        }
        else if ( _last->full() ) {
            _last->_next = std::make_unique< Node >();
            _last->_next->_prev = _last;
            _last = _last->_next.get();
        }
        return _last->push_back( std::forward< TT >( val ) );
    }

    // first is kept using unique_ptr to allow automatic deallocation, last has
    // to be raw pointer to prevent double ownership
    std::unique_ptr< Node > _first;
    Node *_last;
};
